"""
utility functions to make permission handling easier
"""

from django.http import HttpResponseRedirect, Http404

def redirect_to_not_allowed():
    return HttpResponseRedirect('/not_allowed/')

def permission_required(perm):
    """
    Decorator for views that checks whether a user has a particular permission
    enabled, redirecting to a not allowed page if necessary.
    """
    def decorate(view_func):
        def func(request, *args, **kwargs):
            if request.user.has_perm(perm):
                return view_func(request, *args, **kwargs)
            return redirect_to_not_allowed()
        return func
    return decorate

def check_permission(user, permission, raise404=False):
    """
    Utility function to check a permission for a user.
    Redirects to not allowed page if the user is not authorized.
    """
    if not user.has_perm(permission):
        if raise404:
            raise Http404
        return redirect_to_not_allowed()

def check_object_permission(user, object, objectpermission, permfunckwargs={}, raise404=False):
    """
    Utility function to check a permission method on an object.
    Redirects to not allowed page if the method exists and returns false for the user
    If the method does not exist, falls back to a permission check
    """
    object_attribute = 'may_%s' % objectpermission
    if hasattr(object, object_attribute):
        if not getattr(object, object_attribute)(user, **permfunckwargs):
            if raise404:
                raise Http404
            return redirect_to_not_allowed()
    else:
        permission = '%s.%s_%s' % (object._meta.app_label.lower(), objectpermission, object._meta.object_name.lower())
        return check_permission(user, permission, raise404=raise404)

def login_required(function=None):
    """
    Decorator for views that checks whether a user is logged in,
    redirecting to a not allowed page if necessary.
    """
    def decorate(view_func):
        def func(request, *args, **kwargs):
            if request.user.is_authenticated():
                return view_func(request, *args, **kwargs)
            return redirect_to_not_allowed()
        return func
    if function:
        return decorate(function)
    return decorate
