"""
Base implementation of mptt tree model with versioning

Extends the VersionedModel with the mptt methods

"""
from django.db.models import signals as model_signals
from django.db import models
from django.db import connection, transaction

from utils.versioning import models as versioning_models

from mptt.signals import pre_save
from mptt.managers import TreeManager
from mptt import models as mptt_models

qn = connection.ops.quote_name

class VersionedMpttManager(TreeManager):
    """
    Specific manager for versioned mptt models.
    Uses the VersionedModelQuerySet
    """
    def __init__(self):
        models.Manager.__init__(self)
        self.parent_attr = 'parent'
        self.left_attr = 'lft'
        self.right_attr = 'rght'
        self.tree_id_attr = 'tree_id'
        self.level_attr = 'level'

    def get_query_set(self):
        return VersionedMpttModelQuerySet(self.model).order_by(
            self.tree_id_attr, self.left_attr)

    def get_versionedobject(self, version=None, **kwargs):
        qs = self.get_query_set().filter(**kwargs).get_versionedobjects(version)
        if qs:
            object = qs[0]
        else:
            raise self.model.DoesNotExist
        return object

    def get_versionedobjects(self, version=None):
        return self.get_query_set().get_versionedobjects(version)

    # rebuild tree based on patch from here:
    # http://code.google.com/p/django-mptt/issues/detail?id=13
    def rebuild(self):
        """
        Rebuilds whole tree in database using `parent` link.
        """
        opts = self.model._meta

        cursor = connection.cursor()
        cursor.execute('UPDATE %(table)s SET %(left)s = 0, %(right)s = 0, %(level)s = 0, %(tree_id)s = 0' % {
            'table': qn(opts.db_table),
            'left': qn(opts.get_field(self.left_attr).column),
            'right': qn(opts.get_field(self.right_attr).column),
            'level': qn(opts.get_field(self.level_attr).column),
            'tree_id': qn(opts.get_field(self.tree_id_attr).column)
        })

        if hasattr(opts, 'order_insertion_by') and opts.order_insertion_by:
            temp_order_by = 'ORDER BY ' + ', '.join([qn(field) for field in opts.order_insertion_by])
        else:
            temp_order_by = ''
        cursor.execute('SELECT %(id_col)s FROM %(table)s WHERE %(parent_col)s is NULL %(orderby)s' % {
            'id_col': qn(opts.pk.column),
            'table': qn(opts.db_table),
            'parent_col': qn(opts.get_field(self.parent_attr).column),
            'orderby': temp_order_by
        })

        idx = 0
        for (pk, ) in cursor.fetchall():
            idx += 1
            self._rebuild_helper(pk, 1, idx)
        transaction.commit_unless_managed()

    def _rebuild_helper(self, pk, left, tree_id, level=0):
        opts = self.model._meta
        right = left + 1

        if hasattr(opts, 'order_insertion_by') and opts.order_insertion_by:
            temp_order_by = 'ORDER BY ' + ', '.join([qn(field) for field in opts.order_insertion_by])
        else:
            temp_order_by = ''
        cursor = connection.cursor()
        cursor.execute('SELECT %(id_col)s FROM %(table)s WHERE %(parent_col)s = %(parent)d %(orderby)s' % {
            'id_col': qn(opts.pk.column),
            'table': qn(opts.db_table),
            'parent_col': qn(opts.get_field(self.parent_attr).column),
            'parent': pk,
            'orderby': temp_order_by
        })

        for (child_id, ) in cursor.fetchall():
            right = self._rebuild_helper(child_id, right, tree_id, level+1)

        cursor.execute("""
            UPDATE %(table)s
            SET
                %(left_col)s = %(left)d,
                %(right_col)s = %(right)d,
                %(level_col)s = %(level)d,
                %(tree_id_col)s = %(tree_id)d
            WHERE
                %(pk_col)s = %(pk)s
        """ % {
            'table': qn(opts.db_table),
            'pk_col': qn(opts.pk.column),
            'left_col': qn(opts.get_field(self.left_attr).column),
            'right_col': qn(opts.get_field(self.right_attr).column),
            'level_col': qn(opts.get_field(self.level_attr).column),
            'tree_id_col': qn(opts.get_field(self.tree_id_attr).column),
            'pk': pk,
            'left': left,
            'right': right,
            'level': level,
            'tree_id': tree_id
        })

        return right + 1

    def restore_version(self, version):
        """
        generic inefficient restore, which replaces all trees with old version
        in the general case there is no way to link the new and old trees


        Used for installation trees to preserve order of trees
        """
        for item in self.root_nodes().filter(version__isnull=True):
            for child in item.get_children():
                child.mptt_delete()
            item.delete(version.user, nocascade=True)
        for item in self.root_nodes().get_versionedobjects(version).order_by('tree_id'):
            item.restore_version(version)

class VersionedMpttModelQuerySet(versioning_models.VersionedModelQuerySet):
    """
    ! do not use with filter on parent, because this gets overruled here in orde
    to get trees
    """
    def get_versionedobjects(self, version=None):
        roots = versioning_models.VersionedModelQuerySet.get_versionedobjects(self.filter(parent__isnull=True), version)
##        roots = versioning_models.VersionedModelQuerySet.get_versionedobjects(self, version)
        tree_ids = set([item.tree_id for item in roots if not item.deleted])
        return versioning_models.VersionedModelQuerySet.get_versionedobjects(self._clone().filter(tree_id__in=tree_ids), version)

def make_dirty(func):
    """
    Decorator to make tree items dirty when moving.
    """
    def _wrapper(self, user, *args, **kwargs):
        if not self.dirty:
            self.dirty=True
            self.published_dirty = True
            self.save(user)
        return func(self, *args, **kwargs)
    return _wrapper

class MpttVersionedModelManager(versioning_models.VersionedModelManager):
    """
    Subclass of VersionedModelManager
    Replaces get_newerobjects for trees
    """
    def get_newerobjects(self, version=None):
        """
        Custom for trees
        Returns a list, not a queryset
        """
        result = []
        qs = super(MpttVersionedModelManager, self).get_newerobjects(version)
        if qs:
            for item in set([x.tree_id for x in qs]):
                try:
                    el = super(MpttVersionedModelManager, self).filter(tree_id=item)[0]
                    result.append(el)
                except:
                    pass
        return result

    def restore_version(self, version):
        """
        better version, only restoring what was changed
        ! order between trees is not preserved !
        """
        qs = super(MpttVersionedModelManager, self).get_newerobjects(version)
        if qs:
            newitems = []
            olditems = []
            for item in qs:
                try:
                    it = self.get_query_set().get(v_id=item.v_id, version__isnull=True)
                    newitems.append(it)
                except:
                    pass
                try:
                    it = item.get_version(version)
                    if it:
                        olditems.append(it)
                except:
                    pass

            newtrees = set([item.tree_id for item in newitems if item is not None])
            oldtrees = set([item.tree_id for item in olditems if item is not None])

            for item in newtrees:
                el = self.get_query_set().get(tree_id=item, parent__isnull=True)
                for child in el.get_children():
                    child.mptt_delete()
                el.delete(version.user, nocascade=True)
            for item in oldtrees:
                el = self.get_query_set().get(tree_id=item, parent__isnull=True)
                el.restore_version(version)

def VersionedMpttModel_pre_save(instance, user, **kwargs):
    """
    copied from mptt
    needed to modify to insert setting the user
    calling from save instead of using signal because we need to pass in a user
    """
    opts = instance._meta
    parent = getattr(instance, opts.parent_attr)
    if not instance.pk:
        if (getattr(instance, opts.left_attr) and
            getattr(instance, opts.right_attr)):
            # This node has already been set up for insertion.
            return

        if opts.order_insertion_by:
            right_sibling = _get_ordered_insertion_target(instance, parent)
            if right_sibling:
                instance.insert_at(user, right_sibling, 'left')
                return

        # Default insertion
        instance.insert_at(user, parent, position='last-child')
    else:
        # TODO Is it possible to track the original parent so we
        #      don't have to look it up again on each save after the
        #      first?
        old_parent = getattr(instance._default_manager.get(pk=instance.pk),
                             opts.parent_attr)
        if parent != old_parent:
            setattr(instance, opts.parent_attr, old_parent)
            try:
                if opts.order_insertion_by:
                    right_sibling = _get_ordered_insertion_target(instance,
                                                                  parent)
                    if right_sibling:
                        instance.move_to(user, right_sibling, 'left')
                        return

                # Default movement
                instance.move_to(user, parent, position='last-child')
            finally:
                # Make sure the instance's new parent is always
                # restored on the way out in case of errors.
                setattr(instance, opts.parent_attr, parent)

class VersionedMpttModel(versioning_models.VersionedModel):
    """
    Versioned mptt implementation.
    Uses most functionality from django-mptt, combined with some specific
    versioning items.

    Versioning is always done for the entire tree.
    """
    rght = models.PositiveIntegerField(db_index=True, editable=False)
    lft = models.PositiveIntegerField(db_index=True, editable=False)
    tree_id = models.PositiveIntegerField(db_index=True, editable=False)
    level = models.PositiveIntegerField(db_index=True, editable=False)
    parent = models.ForeignKey('self', null=True, blank=True, related_name="%(class)s_children")

    objects = MpttVersionedModelManager()
    tree = VersionedMpttManager()

    class Meta:
        abstract = True

    get_ancestors = mptt_models.get_ancestors
    get_children = mptt_models.get_children
    get_descendants = mptt_models.get_descendants
    get_descendant_count = mptt_models.get_descendant_count
    get_next_sibling = mptt_models.get_next_sibling
    get_previous_sibling = mptt_models.get_previous_sibling
    get_root = mptt_models.get_root
    get_siblings = mptt_models.get_siblings
    is_child_node = mptt_models.is_child_node
    is_leaf_node = mptt_models.is_leaf_node
    is_root_node = mptt_models.is_root_node

    insert_at = make_dirty(mptt_models.insert_at)
    move_to = make_dirty(mptt_models.move_to)

    def save(self, user, set_dirty=True, *args, **kwargs):
        VersionedMpttModel_pre_save(self, user)
        super(VersionedMpttModel, self).save(user, set_dirty=set_dirty, *args, **kwargs)
        # make root dirty, versioning is based on root
        if set_dirty and not self.is_root_node():
            root = self.get_root()
            root.dirty = True
            root.published_dirty = True
            root.save(user)
    save.alters_data = True

    def delete(self, user, override=False):
        if override or self.new:
            self.mptt_delete()
        else:
            # make old root dirty
            if not self.is_root_node():
                root = self.get_root()
                root.dirty = True
                root.published_dirty = True
                root.save(user)
# need to reload items in between here, because otherwise the values for lft and
# rght are wrong
# can this be avoided ??
            parent = self.parent
            for element in self.get_children():
                elem = self.__class__.objects.get(pk=element.pk)
                elem.move_to(user, parent, position='last-child')
                parent = self.__class__.objects.get(pk=parent.pk)
            self = self.__class__.objects.get(pk=self.pk)
            self.deleted = True
            # relocate out of tree
            self.parent = None
            self.save(user)
    delete.alters_data = True

    def mptt_delete(self):
        """
        Delete like needed for a clean mptt model.
        """
        opts = self._meta
        tree_width = self.rght - self.lft + 1
        target_right = self.rght
        tree_id = self.tree_id
        self._tree_manager._close_gap(tree_width, target_right, tree_id)
        super(versioning_models.VersionedModel, self).delete()

    def make_version(self, version, new_tree_id=None, force_version=False):
        """
        Special make_version works recursive and gives versioned items their own tree_id.
        Apply this on the root items of the tree.
        """
        # no action if item is already a versioned item
        toversion = False
        if not self.version:
            if force_version:
                toversion = True
            if self.is_root_node():
                # ignore child nodes except when force_version is set
                if self.dirty:
                    toversion=True
                elif version.published and self.published_dirty:
                    toversion=True
        if toversion:
            a = self.get_version(version, exactversion=True)
            if not a:
                if not new_tree_id:
                    new_tree_id = self._tree_manager._get_next_tree_id()
                # reset dirty flag
                self.dirty = False
                if version.published:
                    self.published_dirty = False
                    self.published = True
                currentpk = self.pk
                currenttree_id = self.tree_id
                # setting pk to None will save a new object in the database
                self.pk = None
                self.tree_id = new_tree_id
                self.version=version
                oldvals = {}
                for field in self._meta.fields:
                    if isinstance(field, models.ForeignKey):
                        oldvals[field] = getattr(self, field.name)
                self.prepare_object_for_versioning(version)

                super(versioning_models.VersionedModel, self).save()

                newpk = self.pk

                # reset pk
                self.pk = currentpk
                self.tree_id = currenttree_id
                if self.deleted:
                    # now really delete the item
                    self.mptt_delete()
                else:
                    # reset version
                    for key, value in oldvals.items():
                        setattr(self, key.name, value)
                    self.version = None
                    self.new = False
                    self.published = False
                    super(versioning_models.VersionedModel, self).save()
                for child in self.get_children():
                    child.make_version(version, new_tree_id, force_version=True)

                return newpk
            else:
                return a.pk
    make_version.alters_data = True

    def restore_version(self, version):
        """
        Special restore_version restores entire tree.
        Apply this on the root items of the tree.
        """
        if not self.deleted:
            if self.version:
                # not using get_version to allow getting deleted objects
                se = self.__class__.objects.filter(v_id=self.v_id, version__isnull=True)
                if se:
                    wvobject = se[0].get_root()
                else:
                    wvobject = None
            else:
                wvobject = self.get_root()

            versionobject = self.get_version(version)

            # delete tree
            if wvobject:
                wvobject.mptt_delete()

            # copy tree from version
            if versionobject:
                versionobject = versionobject.get_root()
                tree = versionobject.get_descendants(include_self=True)
                newtree_id = self.__class__.tree._get_next_tree_id()
                pkdict = {}
                for item in tree:
                    oldv_id = item.v_id
                    item.pk = None
                    item.tree_id = newtree_id
                    item.dirty = True
                    item.published_dirty = True
                    item.published = False
                    item.version = None
                    if item.parent:
                        item.parent = pkdict.get(item.parent.v_id)
                    super(versioning_models.VersionedModel, item).save()
                    pkdict[oldv_id] = item

    restore_version.alters_data = True

    def __init__(self, *args, **kwargs):
        super(VersionedMpttModel, self).__init__(*args, **kwargs)
        opts = self._meta
        opts.parent_attr = 'parent'
        opts.right_attr = 'rght'
        opts.left_attr = 'lft'
        opts.tree_id_attr = 'tree_id'
        opts.level_attr = 'level'
        opts.tree_manager_attr = 'tree'
        opts.order_insertion_by = None
        self._tree_manager = getattr(self.__class__, 'tree')

model_signals.pre_save.connect(pre_save, sender=VersionedMpttModel)
